; ACL2 Version 1.8

; Copyright (C) 1989-95 Computational Logic, Inc. (CLI).  All rights reserved.

; Use of this software constitutes agreement with the terms of ACL2
; license agreement, found in the file LICENSE.

; ACL2 (A Computational Logic for Applicative Common Lisp) was written at
; Computational Logic, Inc., primarily by

;   Matt Kaufmann
;   and
;   J Strother Moore

; Acknowledgements:  Bob Boyer contributed crucially to the design and
; implementation of early versions of ACL2.  Many others, largely at CLI, have
; also contributed to the design of certain features.  We especially thank
; Bishop Brock and John Cowles for their contributions.  See the documentation
; topic ACKNOWLEDGEMENTS for more information.

; This research was supported in part at Computational Logic, Inc. by the
; Defense Advanced Research Projects Agency (DARPA), ARPA Orders #6082, 9151
; and 7406 and by the Office of Naval Research, contracts numbers
; N00014-88-C-0454, N00014-91-C-0130, and N00014-94-C-0193.  The views and
; conclusions contained in this document are those of the author(s) and should
; not be interpreted as representing the official policies, either expressed or
; implied, of Computational Logic, Inc., Office of Naval Research, DARPA or the
; U.S. Government.

; This file, acl2.lisp, (a) builds the packages for the ACL2 system, (b)
; defines the functions for loading and compiling ACL2 and for running code
; verified using ACL2, and (c) makes a couple of checks concerning minor,
; non-CLTL, assumptions that we make about Common Lisps in which ACL2 is to
; run.

;                               FILES


; This is the file acl2.lisp, the first source file for ACL2.  The
; names of the other ACL2 source files are listed under *acl2-files*
; in the file axioms.lisp.

; All of ACL2 is written in Common Lisp and we expect that ACL2 will
; run in any Common Lisp except for those Common Lisps which fail the
; tests we execute upon loading this file, acl2.lisp.  With the
; exception of this file, the file interface-raw.lisp, and those
; constructs after the #-acl2-loop-only passim only (especially in the
; file axioms.lisp), ACL2 is written in the applicative Common Lisp
; for which ACL2 is a verification system.


;                            LISP BUGS

; To use Acl2 under Lispworks 3.2.0, execute the following to workaround a bug.

#+lispworks
(setf (svref ccl::*funny-symbol-char-table* (char-code #\.)) t)


;                            CLTL1/CLTL2

; For the most part, we program in a subset both of CLTL1 and CLTL2.
; However, there are a few places, notably the name of the main
; package for Lisp symbols, where this is impossible.  So we use the
; read conditional #+CLTL2.  If one wishes to run ACL2 under CLTL2,
; execute the following form before commencing compiling or loading:

;        (push :CLTL2 *features*)

; For example, for Allegro and lispworks (respectively) we have the following.

#+(or DRAFT-ANSI-CL-2 lispworks)
(push :CLTL2 *features*)

; The theory will change as a result of this ``push'' in one important
; way:  the package-name of the main Lisp symbols will become
; "COMMON-LISP" instead of "LISP".

; We use IN-PACKAGE in a way that is consistent with both CLTL1 and
; CLTL2, namely as a macro (i.e., whose argument is not evaluated) for
; switching *package* to an existing package, value ignored.

(in-package #-CLTL2 "USER" #+CLTL2 "COMMON-LISP-USER")


;                              PACKAGES

; We never intend that this file should be compiled, and hence we do
; not need to obey the CLTL1 strictures about putting all IN-PACKAGE
; forms near the front.

(eval-when #-cltl2 (compile)
           #+cltl2 (:compile-toplevel)
           (error "The file acl2.lisp should never be compiled."))

(dolist
 (p (list-all-packages))
 (cond ((equal 3 (string< "ACL" (package-name p)))
        (format t
                "~%~%Warning:  There is already a package with the ~
                 name ~a. ~%The ACL2 implementation depends upon ~
                 having complete knowledge of ~%and control over any ~
                 packge whose name begins with the ~%three letters ``ACL'', ~
                 so ACL2 may not work in this Lisp." (package-name p))
        (cond ((package-use-list p)
               (format t "~%~%Warning:  The package with name ~a ~
                   USES the packages in the list ~a.  ACL2 will not work ~
                   in state of affairs."
                       (package-name p) (package-use-list p)))))))

(or (find-package "ACL2")
    (make-package "ACL2" :use nil))

(defmacro acl2::defconst (&rest acl2::x)
  (cons 'defparameter acl2::x))

; We now get our imports for package ACL2, putting them into the
; variable acl2::*common-lisp-symbols-from-main-lisp-package*.

; We use different variables for common-lisp-symbols-from-main-lisp-package*,
; *acl2-version*, and *acl2-files*, in order to avoid compiler warnings for
; redefined variables.

(defvar acl2::*copy-of-common-lisp-symbols-from-main-lisp-package*)
(defvar acl2::*copy-of-common-lisp-specials-and-constants*)
(defvar acl2::*copy-of-acl2-version*)
(defvar acl2::*copy-of-acl2-files*)

(with-open-file
 (fl (open "axioms.lisp" :direction :input))

;  Get into the main lisp package in order to read in
;  *common-lisp-symbols-from-main-lisp-package*.

 (let ((*package* (find-package #-CLTL2 "LISP" #+CLTL2 "COMMON-LISP")))

;  Skip the in-package

   (read fl)

;  Do the defconst for *common-lisp-symbols-from-main-lisp-package*.

   (setq acl2::*copy-of-common-lisp-symbols-from-main-lisp-package*
         (eval (caddr (read fl))))
   (import acl2::*copy-of-common-lisp-symbols-from-main-lisp-package* "ACL2")

;  Do the defconst for *acl2-version*.

   (setq acl2::*copy-of-acl2-version*
         (eval (caddr (read fl))))

;  Do the defconst for *acl2-files*.

   (setq acl2::*copy-of-acl2-files*
         (eval (caddr (read fl))))

;  Do the defconst for *common-lisp-specials-and-constants*.

   (setq acl2::*copy-of-common-lisp-specials-and-constants*
         (eval (caddr (read fl))))))

(in-package "ACL2")

(defparameter *global-package-prefix* "ACL2_GLOBAL_")

(defparameter *1*-package-prefix* "ACL2_*1*_")

(defun make-global-package (x)
  (let* ((n (concatenate 'string *global-package-prefix* x))
         (p (find-package n)))
    (cond (p
           (do-symbols (sym p)
                       (makunbound sym)))
          (t (make-package n :use nil)))))

(defun make-*1*-package (x)
  (let* ((n (concatenate 'string *1*-package-prefix* x)))

; Unlike make-global-package, here we really don't have to worry about bound
; (or fbound) symbols.  Presumably ``bound'' is irrelevant, and ``fbound'' is
; taken care of by our undoing mechanisms.

    (make-package n :use nil)))

(or (find-package "ACL2-INPUT-CHANNEL")
    (make-package "ACL2-INPUT-CHANNEL" :use nil))

(or (find-package "ACL2-OUTPUT-CHANNEL")
    (make-package "ACL2-OUTPUT-CHANNEL" :use nil))


; Here are the initial global and *1* packages.

; Keep the following in sync with *main-lisp-package-name*.
(make-global-package #-CLTL2 "LISP" #+CLTL2 "COMMON-LISP")
(make-global-package "KEYWORD")
(make-global-package "ACL2")
(make-global-package "ACL2-INPUT-CHANNEL")
(make-global-package "ACL2-OUTPUT-CHANNEL")

; Keep the following in sync with *main-lisp-package-name*.
(make-*1*-package #-CLTL2 "LISP" #+CLTL2 "COMMON-LISP")
; Functions cannot be defined in the keyword package, so we do not do so.
;  (make-*1*-package "KEYWORD")
(make-*1*-package "ACL2")
(make-*1*-package "ACL2-INPUT-CHANNEL")
(make-*1*-package "ACL2-OUTPUT-CHANNEL")

; The ACL2_INVISIBLE packages is used for a few odd objects that are
; to be hidden from the ACL2 user.

(let ((inv "ACL2_INVISIBLE"))
  (or (find-package inv)
      (make-package inv :use nil)))

; Common Lisp does not require that the symbol-package of the basic
; Lisp symbols be the basic LISP or COMMON-LISP package, but merely
; that those symbols be exported from that package.

(defparameter acl2::*initial-lisp-symbol-mark*
  'acl2_invisible::initial-lisp-symbol-mark)

(dolist (sym *copy-of-common-lisp-symbols-from-main-lisp-package*)
        (setf (get sym *initial-lisp-symbol-mark*)
              #-CLTL2 "LISP"
              #+CLTL2 "COMMON-LISP"))

(defconstant *acl2-package* (find-package "ACL2"))

(dolist (x *features*)
        (cond ((or (equal "AKCL-SET-MV"
                          (symbol-name x))
                   (equal "ACL2-LOOP-ONLY"
                          (symbol-name x))
                   (equal "ACL2-METERING"
                          (symbol-name x)))
               (format t "~%~%Warning:  This Common Lisp may be ~
                          unsuitable for ACL2 because a symbol with~
                          ~%the name \"ACL2-LOOP-ONLY\", ~
                          \"AKCL-SET-MV\" or \"ACL2-METERING\" is ~
                          a member of *FEATURES*."))))

#+akcl
(if (fboundp 'si::set-mv)
    (push :akcl-set-mv *features*)
  (error "Use a version of ACKL after 206"))


;                                CHECKS

(defconstant acl2::*acl2-status-file*
  (make-pathname :name "acl2-status"
                 :type "txt"))

(defun acl2::check-suitability-for-acl2 ()
  (progn (or (not (probe-file *acl2-status-file*))
             (delete-file *acl2-status-file*))
         (load "acl2-check.lisp")
         (with-open-file (str *acl2-status-file*
                              :direction :output)
                         (format str
                                 "~s"
                                 :checked))
         t))

(defun note-compile-ok ()
  (progn (or (not (probe-file *acl2-status-file*))
             (delete-file *acl2-status-file*))
         (with-open-file (str *acl2-status-file*
                              :direction :output)
                         (format str
                                 "~s"
                                 :compiled))))


;                       COMPILING and LOADING

; To compile ACL2, invoke (load "acl2.lisp") and then invoke
; (acl2::compile-acl2).  Having compiled ACL2, to build ACL2, invoke
; (load "acl2.lisp") and then invoke (acl2::load-acl2).  To run ACL2
; verified code, one may either load all of ACL2 as above, or
; alternatively, having complied ACL2, invoke (load "acl2.lisp") and
; invoke (acl2::load-acl2-execution-environment).  The top-level user
; functions for ACL2 are in package "ACL2", so invoke (in-package
; "ACL2") to obviate typing package names.

; NOTE:  In order to compile ACL2, checks must first be run on the suitability
; of the underlying Common Lisp implementation, by executing
; (check-suitability-for-acl2).  If the Common Lisp is suitable, this form will
; write the file acl2-status.txt with the symbol :CHECKED.  Successful
; compilation should write out that same file with the symbol :COMPILED.

; The detailed setting of the ACL2 readtable occurs later in this
; file, but we put the declaration of this special before its first
; use.

(defparameter *acl2-readtable*
  (copy-readtable nil)
  "*acl2-readtable* is the readtable we use (a) to prevent the use
#. to cause evaluation during READing (b) and to define our own version
of backquote.")

(defvar *lisp-extension* "lisp")

(defparameter *compiled-file-extension*
  #+akcl "o"
  #+kcl  "o"
  #+(and lucid sparc) "sbin"
  #+(and lucid mc68000) "lbin"
  #+allegro "fasl"
  #+lispworks "wfasl"
  #+mcl "fasl"
  #+symbolics "bin"
  #-(or akcl kcl lucid allegro lispworks mcl symbolics)
  (do ((name
        (progn
          (format
           t
           "~%~%We are trying to initialize the parameter that specifies the~%~
            file extension name used by the local compiler.  For example, in~%~
            Allegro Common Lisp, the extension is \"fasl\" while in Lucid~%~
            it is \"lbin\" or \"sbin\".  Files with such extensions are understood~%~
            to contain compiled code.  We do not recognize this Common Lisp~%~
            from its *features*.  Please type as a STRINGP suitable for READ~%~
            its compiled file extension name:  ")
          (read t))
        (progn
          (format
           t
           "~%~%It has to READ as a STRINGP!  Put quotation marks ~
            around it:  ")
          (read t))))
      ((stringp name) name)))

; The user is welcome to modify the following proclaim form.

(proclaim (quote (optimize #+(or cmu lucid lispworks) (compilation-speed 0)
                           ;; #+allegro (debug 0)
                           #+cmu (extensions:inhibit-warnings 3)
                           (speed 3)
                           (space #+cmu 1 #-cmu 0)
                           (safety 0))))

#+allegro
(eval (read-from-string "
  (SETQ COMPILER::DECLARED-FIXNUMS-REMAIN-FIXNUMS-SWITCH
        #'(LAMBDA (X Y Z #+(VERSION>= 4 1) D) NIL)) "))

(defmacro our-with-compilation-unit (form)

; In fact, with-compilation-unit is only defined in dpANS, not CLtL2.  But MCL
; does seem to support it, so we allow it with #+cltl2.

  #+cltl2 
  `(common-lisp::with-compilation-unit () ,form)
  #-cltl2
  form)

(defun maybe-push-small-acl2-image-onto-features (small-p)
  (cond
   ((and small-p
         (not (member :small-acl2-image *features*)))
    (push :small-acl2-image *features*))))

(defun compile-acl2 (&optional (small-p t))

; Here is a natural place to put compiler options.  In fact, we put
; them above, globally.

; (proclaim '(optimize (safety 0) (space 0) (speed 3)))

; Note on Illegal Instructions:  If ACL2 causes an illegal instruction
; trap it is difficult to figure out what is happening.  Here is a
; brute force way to do it.
; (0) create an events file or book that will recreate the
;     the state up to the event that causes the illegal instruction.
; (1) Copy saved_acl2 to old_saved_acl2
; (2) replace the proclaim above with
;     (proclaim '(optimize (safety 3) (space 0) (speed 0)));
; (3) make full init
;     this will fail because of stack overflows during booting.
; (4) fire up old_saved_acl2
; (5) load the saved events file or book, recreating the state.
; (6) :q from (LP)
; (7) (progn (setq old-world-pair (get 'current-acl2-world 'acl2-world-pair))
;            t)
; (8) (dolist (name *copy-of-acl2-files*)
;           (or (equal name "defpkgs")
;               (load source)))   ; load the (safety 3) .o files
;     Actually, the last time I did this I loaded each file manually
;     and I did not load ld or interface-raw.  I think they can be
;     loaded here though.
; (9) (progn (f-put-global 'current-acl2-world
;                          (car old-world-pair) *the-live-state*)
;            t)
;(10) (si::use-fast-links nil)
;(11) enter (LP) and execute the offending event.
;(12) once the bug is fixed, be sure to change the proclaim
;     back to (safety 0) (speed 3) before recompiling.

; Note on loading before compiling.  We load each ACL2 source file
; before compiling it to make sure that the functions needed to
; execute macros have been defun-ed before they are called.  Normal
; Common Lisp compilation does not do this.  So we cause all forms to
; be executed before we start the compilation.  This guarantees that
; when macros run, the functions they call have been defined.

; In general, and for the same reason, all ACL2 user checked files
; are also be loaded before they are compiled.

  (maybe-push-small-acl2-image-onto-features small-p)
  (cond
   ((or (not (probe-file *acl2-status-file*))
        (with-open-file (str *acl2-status-file*
                             :direction :input)
                        (not (eq (read str nil)
                                 :checked))))
    (check-suitability-for-acl2)))
  (our-with-compilation-unit
   (let ((*readtable* *acl2-readtable*)
         #+akcl

; AKCL compiler note stuff.  We have so many tail recursive functions
; that the notes about tail recursion optimization are just too much
; to take.

         (compiler:*suppress-compiler-notes* t))
     (dolist (name *copy-of-acl2-files*)
             (or (equal name "defpkgs")
                 (let ((source (make-pathname :name name
                                              :type *lisp-extension*)))
                   (load source)
                   (or (equal name "proof-checker-pkg")
                       (progn
                         (proclaim-file source)
                         (compile-file source)
                         (load (make-pathname :name name
                                              :type *compiled-file-extension*)))))))
; ?? (compare-check-sums)
     (note-compile-ok)
     )))

(defun load-acl2 (&optional (small-p t))
  (maybe-push-small-acl2-image-onto-features small-p)
  #+akcl

; We grow the image slowly even when not building a small image, since we now
; do allocation on start-up.  We are assuming that people will be using
; load-acl2 only as part of the process of building a saved image, and hence
; that this slow growth policy will be undone by the function
; save-acl2-in-akcl.

  (sloop::sloop
   for type in
   '(cons fixnum symbol array string cfun sfun

; In akcl, at least some versions of it, we cannot call allocate-growth on the
; following two types.

          #+gcl contiguous
          #+gcl relocatable
          )
   do
   (cond
    ((or (boundp 'si::*gcl-major-version*) ;GCL 2.0 or greater
         (and (boundp 'si::*gcl-version*) ;GCL 1.1
              (= si::*gcl-version* 1)))
     (si::allocate-growth type 1 10 50 2))
    (t (si::allocate-growth type 1 10 50))))
  (cond
   ((or (not (probe-file *acl2-status-file*))
        (with-open-file (str *acl2-status-file*
                             :direction :input)
                        (not (eq (read str nil)
                                 :compiled))))
    (error "Please compile ACL2 using ~s or~%~
            ~s, which will write the~%~
            token :compiled to the file acl2-status.txt."
           (if small-p '(compile-acl2 t) '(compile-acl2))
           (if small-p
               '(quick-compile-acl2 <very-fast-p> t)
             '(quick-compile-acl2 &optional very-fast)))))
  (let ((*readtable* *acl2-readtable*))
    (dolist (name *copy-of-acl2-files*)
      (or (equal name "defpkgs")
          (if (equal name "proof-checker-pkg")
              (load "proof-checker-pkg.lisp")
            (load (make-pathname :name name
                                 :type *compiled-file-extension*)))))
    (load "defpkgs.lisp")
    (in-package "ACL2")

; Do not make state special, as that can interfere with tail recursion removal.
; The following form is provided merely as a convenience to the user, who may
; want to execute ACL2 forms in raw Lisp.  The eval form is there as an attempt
; to keep Lisp systems from deciding that state should be special upon reading
; this file.

    (eval '(setq state *the-live-state*))
    "ACL2"))

(defun load-acl2-execution-environment (&optional (small-p t))
  (maybe-push-small-acl2-image-onto-features small-p)
  (let ((*readtable* *acl2-readtable*))
    (load (make-pathname :name "axioms"
                         :type *compiled-file-extension*))
    (load "defpkgs.lisp")
    (in-package "ACL2")
    "ACL2"))

;                           BASIC FUNCTIONS TO BE COMPILED

;  Functions for proclaiming and for defining ACL2's Implementation of the
;  Backquote Readmacro are to be compiled, so we do not include them in the
;  present file.  Instead:

(let ((object-file (make-pathname :name "acl2-fns"
                                  :type *compiled-file-extension*)))
  (cond
   ((and (probe-file object-file)
         (<= (file-write-date "acl2-fns.lisp")
             (file-write-date object-file)))
    (load object-file))
   (t
    (load "acl2-fns.lisp") ;we like to load before compiling
    (compile-file "acl2-fns.lisp")
    (load object-file))))

;                           ACL2-READTABLE

;  We now make all of the modifications to *acl-readtable*.

(eval-when
 #-cltl2
 (load eval compile)
 #+cltl2
 (:load-toplevel :execute :compile-toplevel)
 (let ((*readtable* *acl2-readtable*))

; Backquote

   (set-macro-character
    #\`
    #'(lambda (stream char)
        (declare (ignore char))
        (let ((*backquote-counter* (1+ *backquote-counter*)))
          (backquote (read stream t nil t)))))

; Comma

   (set-macro-character
    #\,
    #'(lambda (stream char)
        (declare (ignore char))
        (let ((*backquote-counter* (1- *backquote-counter*)))
          (cond ((< *backquote-counter* 0)
                 (clear-input stream)
                 (error "Illegal comma encountered by READ.")))
          (case (peek-char nil stream t nil t)
                (#\@ (read-char stream t nil t)
                 (list *comma-atsign* (read stream t nil t)))
                (#\. (error ",. not allowed in ACL2 backquote forms."))
                (otherwise (list *comma* (read stream t nil t)))))))

;  Turn off #.

;  Actually, we do not turn it off but make it the all-powerful abort signal.
;  That is, if at any time the user types #. while under the *acl2-readtable*
;  the effect is to abort all the way out to the top-most call of LD.  See LD-FN
;  for some comments.

   (set-dispatch-macro-character
    #\#
    #\.
    #'(lambda (stream char n)

; MCL noticed that stream, char, and n are not used:

        (declare (ignore stream char n))

; We used to cause the following hard error in response to #. so pre-empting
; its meaning for abort is no real loss to the ACL2 user.
;       (error "ACL2 does not support #. syntax.")

        (throw 'local-top-level :abort)))

;  Turn off #,

   (set-dispatch-macro-character
    #\#
    #\,
    #'(lambda (stream char n)

; MCL noticed that stream, char, and n are not used:

        (declare (ignore stream char n))
        (error "ACL2 does not support #, syntax.")))

;  Keep control of character reader.

   (set-dispatch-macro-character
    #\#
    #\\
    #'acl2-character-reader)))


#|
                      Remarks on *acl2-readtable*


Because read-object$ calls the Common Lisp function read, read-object$
is a function of the values of the Common Lisp symbols (a)
*readtable*, (b) *package*, and (c) *features*.  In ACL2, the user can
specify the package to use, via in-package, which sets the global
current-package.  The user of ACL2 has no (legitimate) control over the
readtable, which is set above and discussed below.

As for *features*, we currently permit full use of the *features*
facility of the Common Lisp reader, with this caveat: it is an
official part of the ACL2 story that *features* should have the same
setting throughout all phases of ACL2 usage.  That is, the user must
set up *features* at the beginning, before starting to use ACL2 and
the user must then leave *features* alone (even though the
implementors of ACL2 put :acl2-loop-only onto *features* during
boot-strapping).  One bad consequence of our *features* policy is that
verified files will not in general be verifiable or useable in other
Lisp implementations or installations unless the settings of
*features* relevant to one's usages of the #+ and #- are the same in
the two Lisp implementations.  One simple way to obtain peace of mind
on this topic is for the user not to use #+ or #- at all!  It might be
cleaner for us simply to prohibit the use of the #+ and #- readmacros
all together in ACL2.  This could be done at the cost of replacing the
many uses of #+ and #- in axioms.lisp, and a few other places, with
some sort of regular macro.

Below is a detailed examination of the default Common Lisp readtable
from the perspective of ACL2.  We want to make sure that when we read,
we do not have side effects (e.g. via #.) of ACL2 but will merely
either (a) cause an error or (b) generate a Lisp object, which we then
will check with bad-lisp-object before handing it to ACL2 functions.

All of the standard Common Lisp characters are either white space or
constituent, with the following exceptions, which are read macros:

  "  quote
  #  number sign
  '  quote
  () parentheses
  ,  comma
  ;  comment
  \  single escape
  `  backquote
  |  multiple escape

With the exception of number sign, backquote and comma, we certainly
want all of these readmacros to have their usual Common Lisp
semantics.

We have defined our own backquote and discussed it above.

We now review the # situation:

  ## and #= are for reading possibly circular stuff
         bad-lisp-object may run forever
  #'  reads as function
         won't hurt anything
  #(  vector
         will be rejected by bad-lisp-object
  #)  signals an error
         enough said
  #*  bit vector
         will be rejected by bad-lisp-object
  #,  load-time evaluation
         we shut it off
  #0-#9 used for infix arguments
         ok
  #:  uninterned symbol
         will be rejected by bad-lisp-object
  #<  signals an error
         enough said
  #\  character object
         will be checked by bad-lisp-object; see also below
  #|  start comment, ended by |#
         ok
  #<backspace | tab | newline | linefeed | page | return | space>
      signals an error -- ok
  #+  and #-
      see the discussion of *features* above
  #.  read time evaluation
         we shut it off
  #A, #a  arrays
         will be checked by bad-lisp-object
  #B, #b  binary rational
         ok
  #C, #c complex
         ok (rejected by bad-lisp-object except for rational components)
  #O, #o octal
         ok
  #P, #p pathname
         will be checked by bad-lisp-object
  #R, #r radix-n
         fine
  #S, #s structure
         will be rejected by bad-lisp-object
  #X, #x hex
         ok

Eventually, it will be best to define a read function for ACL2 solely
in terms of ACL2 character reading primitives.  Common Lisp read is
ambiguous.  There is the ambiguity of backquote described above.
There is the abiguity of which tokens get read as numbers.  To make
matters a little more scary, there is nothing that prevents a Common
Lisp implementation from adding, for example, a new # readmacro option
that would provide something as potentially catastrophic as #.  One
obstacle to doing a read within ACL2 this is efficiency.  For example,
ACL2 does not now support unread-char.  And given the requirement that
whatever is specified in a guard must be checkable, it is unclear now
how best to add unread-char since Common Lisp does permit one to
detect whether a stream is in a ``just unread'' state.  ACL2 could
enable one to answer such a question, but at the cost of having to
store the information every time that a character was unread.

|#

;          ACL2's Implementation of the character reader

; We have decided to take full charge of everything that # reader
; does, which is just a part of the way towards writing READ totally
; from scratch.  And we are pretty conservative about what #\ does
; accept; one can always get the characters one wants by using
; code-char.  Notice for example that if we're not careful, then ACL2
; could be potentially unsound when we have 8-bit characters, because
; it's conceivable that
; 
; (equal (char-code "#\fifth-control-character") 5)
; 
; is a theorem in one Common Lisp and yet
; 
; (equal (char-code "#\fifth-control-character") 6)
; 
; is a theorem in another.  Bizarre, but apparently not ruled out by
; dpANS.
; 
; So, we manage this simply by modifying the character reader so that
; the #\ notation only works for single characters and for Space, Tab,
; Newline, Page, and Rubout; an error is caused otherwise.

; Our algorithm for reading character objects starting with #\ is
; quite simple.  We accumulate characters until encountering a
; whitespace character or a terminating macro character, from the list
; *acl2-read-character-terminators*.  The result must be either a
; single standard character or else one of the names (up to case,
; which we ignore in the multiple-character case) SPACE, TAB, NEWLINE,
; PAGE, and RUBOUT.  Otherwise we cause an error.  Note that if we do
; NOT cause an error, then any dpANS-compliant Common Lisp
; implementation's character reader would behave the same way, because
; dpANS says (in the section ``Sharpsign Backslash'') the following.

;    .....  After #\ is read, the reader backs up
;    over the slash and then reads a token, treating the initial slash
;    as a single escape character (whether it really is or not in the
;    current readtable).

; The rather involved description from dpANS in the section ``Reader
; Algorithm'' implies that when a token is terminated without error,
; it must be upon first encountering a whitespace character or a
; terminating macro character.

; Finally, here is an argument that we cannot reasonably allow ACL2 to
; accept character notations of the sort akcl allows, such as #\\112
; for #\J for example.  (By the way, 112 is octal for 74, which is
; (char-code #\J).)  This is sad, because it would have been nice to
; provide a way of reading arbitrary 8-bit characters in ACL2.

; In the following, we refer to documentation from Bill Schelter's
; info version of dpANS.

; So, assume that #\J parses the same as #\\112 (and we'll derive a
; ``contradiction'' of sorts).  The documentation from ``Sharpsign
; Backslash'' below implies that #\\112 parses as the character whose
; name is [STRING-UPCASE applied to] the 4-character string \112.  So
; if #\\112 parses as #\J, then the name of the character #\J is
; "\\112".  Then, the documentation for ``char-name'' (part of which
; is also below) implies that CHAR-NAME returns the character name,
; and hence (CHAR-NAME #\J) must be "\\112".  But probably this isn't
; true of the implementation (it's not true for akcl or allegro, for
; example).  And, it seems really dangerous for us to redefine
; CHAR-NAME in ACL2.

; What's worse, if we apply the first part of this argument to
; #\Newline, we see that the name of the character #\Newline is
; "\\12", which directly contradicts the final passage below from
; dpANS.

; In what follows we'll quote from dpANS (an emacs Info version).  We
; quote here from three sections, separated by ++++++++++++++++++.

#|

Sharpsign Backslash
...................

Syntax: #\<<x>>

[[[ text omitted ]]]

When the token x is more than one character long, the x must have the
syntax of a symbol with no embedded package markers.  In this case, the
sharpsign backslash notation parses as the character whose name is
(string-upcase x); see *See Character Names::.

++++++++++++++++++

char-name                                                        [Function]
---------------------------------------------------------------------------

`char-name'  character =>  name

Arguments and Values::
......................

character--a character.

name--a string or nil.

Description::
.............

Returns a string that is the name of the character, or nil if the
character has no name.

++++++++++++++++++

Character Names
---------------

The following character names must be present in all conforming
implementations:

Newline

|#

;                            DECLARATIONS

; We use XARGS in DECLARE forms.  By making this proclamation, we
; suppress compiler warnings.

(proclaim '(declaration xargs))

